//Terrain class implementation

#include "main.h"

//Abstract terrain class implementation

//RAW map loading
bool TERRAIN::LoadRAWMap(char * file_name, int _size)
{

	FILE* _file = NULL; //pointer to file for loading in

	//check if data was already loaded
	if(m_heightData.m_Data)
	{
		UnloadRAWMap();
	}

	//Open the data file
	_file = fopen(file_name, "rb");
	if(_file == NULL)
	{
		MessageBox(NULL, "Could not open RAW height data file", "ERROR", MB_OK);	
		return false;	
	}

	//allocate memory for RAW file data
	m_heightData.m_Data = new unsigned char[_size * _size * 2];

	//if not...
	if(m_heightData.m_Data == NULL)
	{
		MessageBox(NULL, "Could allocate memory for RAW height data file", "ERROR", MB_OK);	
		return false;		
	}

	//read in the data
	fread(m_heightData.m_Data, _size * _size, 1, _file);

	//close the file
	fclose(_file);

	//Set the size data
	m_heightData.m_Size_ = _size;
	m_Size = m_heightData.m_Size_;

	return true;
}

//RAW map unloading
bool TERRAIN::UnloadRAWMap()
{
	//check if data has been already loaded
	if(m_heightData.m_Data)
	{
		//delete the data
		delete[] m_heightData.m_Data;

		//reset the map dimensions
		m_heightData.m_Size_ = 0;
	}

	return true;
}

////////////////////////////////////////
//Terrain rendering pre-calculations
///////////////////////////////////////
void DEFAULT_TERRAIN::ComputeNormals()
{
	int normalIndex = 0;

	for (int z = -1; z <= m_Size - 1; ++z)
	{
		for (int x = -1; x <= m_Size - 1; ++x)
		{
			// Set/Compute the normal for triangle 1

			float height0 = GetHeightScale( x,    z );
			float height1 = GetHeightScale( x,  z+1 );
			float height2 = GetHeightScale( x+1,  z );
			Vector3 normal1( height0 - height2, 1.0f, height0 - height1 );

			// Set/Compute the normal for triangle 2
			height0 = height2;
			height2 = GetHeightScale( x+1, z+1 );
			Vector3 normal2( height1 - height2, 1.0f, height0 - height2 );

			
			normal1 = normal1.Normalize(normal1);
			normal2 = normal2.Normalize(normal2);


			// Set the surface normals in the array
			m_TerrainNormals[normalIndex++] = normal1;
			m_TerrainNormals[normalIndex++] = normal2;

		}
	}
}

void DEFAULT_TERRAIN::ComputeVertexNormals()
{
	Vector3 vec[6];
	Vector3 vertexNormal;
	int triIndex     = 1;
	int indexPlusOne = 2;
	int indexPlusTwo = 3;
	int vertIndex    = 0;

	// Computes the offset based on the number of triangles across
	int rowOffset = ( ( m_Size + 1 ) * 2 ) - 1;
	for( int z = 0; z < m_Size; z++ )
	{
		for( INT x = 0; x < m_Size; x++ )
		{
			indexPlusOne = triIndex + 1;
			indexPlusTwo = triIndex + 2;

			// Get the three triangles below the vertex
			vec[0] = m_TerrainNormals[ triIndex ];
			vec[1] = m_TerrainNormals[ indexPlusOne ];
			vec[2] = m_TerrainNormals[ indexPlusTwo ];

			// Get the three triangles above the vertex
			vec[3] = m_TerrainNormals[ rowOffset + triIndex ];
			vec[4] = m_TerrainNormals[ rowOffset + indexPlusOne ];
			vec[5] = m_TerrainNormals[ rowOffset + indexPlusTwo ];

			// Sum the vectors and then divide by 6 to average them
			vertexNormal = (vec[0] + vec[1] + vec[2] + vec[3] + vec[4] + vec[5]) / 6;
			
			vertexNormal = vertexNormal.Normalize(vertexNormal);

			// Assign the normals to the stored vertex normals
			m_TerrainNormals[vertIndex] = vertexNormal;

			// Increment the triangle and vertex indices
			triIndex += 2;
			vertIndex++;
		}
		triIndex += 2;
	}
}

//Create the skybox display list
//Consider trying this with a display list, but create it using GL_COMPILE_AND_EXECUTE instead
void DEFAULT_TERRAIN::CreateSkyBox()
{
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	glDisable(GL_CULL_FACE);	

	glPushMatrix();

	glLoadIdentity();

	//The tricky bit...
	//First, make sure the camera is always looking at the skybox
	g_Camera.Look();

	//Then  translate the skybox around the current camera position
	//Disabling the depth testing will then make the skybox seem infinately large (see below in DrawScene())
	glTranslatef(g_Camera.GetPosX(), g_Camera.GetPosY(), g_Camera.GetPosZ());

	//Set the skybox color to white
	glColor3f(255.0f, 255.0f, 255.0f);

	//Skybox front
	glBindTexture(GL_TEXTURE_2D, Textures[1]);
	glBegin (GL_QUADS);
	{
		//Front face
		glNormal3f( 0.0f, 0.0f, 1.f);					
		glTexCoord2f(0.0f, 0.0f);
		glVertex3f(-10.0f, -10.0f,  10.0f);
		glTexCoord2f(1.0f, 0.0f); 
		glVertex3f( 10.0f, -10.0f,  10.0f);
		glTexCoord2f(1.0f, 1.0f); 
		glVertex3f( 10.0f,  10.0f,  10.0f);
		glTexCoord2f(0.0f, 1.0f); 
		glVertex3f(-10.0f,  10.0f,  10.0f);
	}
	glEnd();

	//Skybox back
	glBindTexture(GL_TEXTURE_2D, Textures[0]);
	glBegin (GL_QUADS);
	{
		glNormal3f( 0.0f, 0.0f,-1.f);					
		glTexCoord2f(1.0f, 0.0f); 
		glVertex3f(-10.0f, -10.0f, -10.0f);
		glTexCoord2f(1.0f, 1.0f); 
		glVertex3f(-10.0f,  10.0f, -10.0f);
		glTexCoord2f(0.0f, 1.0f); 
		glVertex3f( 10.0f,  10.0f, -10.0f);
		glTexCoord2f(0.0f, 0.0f); 
		glVertex3f( 10.0f, -10.0f, -10.0f);
	}
	glEnd();
	
	//Skybox top
	glBindTexture(GL_TEXTURE_2D, Textures[3]);
	glBegin (GL_QUADS);
	{
		glNormal3f( 0.0f, 1.f, 0.0f);					
		glTexCoord2f(0.0f, 1.0f); 
		glVertex3f(-10.0f,  10.0f, -10.0f);
		glTexCoord2f(0.0f, 0.0f); 
		glVertex3f(-10.0f,  10.0f,  10.0f);
		glTexCoord2f(1.0f, 0.0f); 
		glVertex3f( 10.0f,  10.0f,  10.0f);
		glTexCoord2f(1.0f, 1.0f); 
		glVertex3f( 10.0f,  10.0f, -10.0f);
	}
	glEnd();

	//Skybox bottom
	glBindTexture(GL_TEXTURE_2D, Textures[2]);
	glBegin (GL_QUADS);
	{
		glNormal3f( 0.0f,-1.f, 0.0f);					
		glTexCoord2f(1.0f, 1.0f); 
		glVertex3f(-10.0f, -10.0f, -10.0f);
		glTexCoord2f(0.0f, 1.0f); 
		glVertex3f( 10.0f, -10.0f, -10.0f);
		glTexCoord2f(0.0f, 0.0f); 
		glVertex3f( 10.0f, -10.0f,  10.0f);
		glTexCoord2f(1.0f, 0.0f); 
		glVertex3f(-10.0f, -10.0f,  10.0f);
	}
	glEnd();

	//Skybox right face
	glBindTexture(GL_TEXTURE_2D, Textures[5]);
	glBegin (GL_QUADS);
	{
		glNormal3f( 1.f, 0.0f, 0.0f);					
		glTexCoord2f(1.0f, 0.0f);
		glVertex3f( 10.0f, -10.0f, -10.0f);
		glTexCoord2f(1.0f, 1.0f); 
		glVertex3f( 10.0f,  10.0f, -10.0f);
		glTexCoord2f(0.0f, 1.0f); 
		glVertex3f( 10.0f,  10.0f,  10.0f);
		glTexCoord2f(0.0f, 0.0f); 
		glVertex3f( 10.0f, -10.0f,  10.0f);
	}
	glEnd();
		
	//Skybox left face
	glBindTexture(GL_TEXTURE_2D, Textures[4]);
	glBegin (GL_QUADS);
	{
		glNormal3f(-1.f, 0.0f, 0.0f);					
		glTexCoord2f(0.0f, 0.0f); 
		glVertex3f(-10.0f, -10.0f, -10.0f);
		glTexCoord2f(1.0f, 0.0f); 
		glVertex3f(-10.0f, -10.0f,  10.0f);
		glTexCoord2f(1.0f, 1.0f); 
		glVertex3f(-10.0f,  10.0f,  10.0f);
		glTexCoord2f(0.0f, 1.0f); 
		glVertex3f(-10.0f,  10.0f, -10.0f);
	}

	glEnd();//end drawing

	glPopMatrix();

	//glDisable(GL_TEXTURE_2D);//temp, remove once terrain gets textured
	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
}

//Draw the water along with the animated texture
void DEFAULT_TERRAIN::CreateWater()
{
	GLfloat WaterDiffuse[] = {0.8f, 0.5f, 0.5f, 0.7f};
	GLfloat WaterSpec[] = {0.8f, 0.8f, 0.8f};
	GLfloat WaterShine[] = {0.8f, 0.8f, 0.8f};

	glPushMatrix();
	{
		//enable sphere mapping
		glTexGeni(GL_S, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
		glTexGeni(GL_T, GL_TEXTURE_GEN_MODE, GL_SPHERE_MAP);
		glEnable(GL_TEXTURE_GEN_S);
		glEnable(GL_TEXTURE_GEN_T);

		//perform alpha blending
		glDisable(GL_CULL_FACE);
		glDepthMask(GL_FALSE);
		glEnable(GL_BLEND);
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

		//bind the water texture
		glBindTexture(GL_TEXTURE_2D, Textures[7]);

		//draw the water body in the middle
		glTranslatef(200.0f, 18.0f, 0.0f);

		//draw the water plane
		glBegin (GL_QUADS);
		{
			glMaterialfv( GL_FRONT, GL_DIFFUSE, WaterDiffuse);
			glMaterialfv( GL_FRONT, GL_SPECULAR, WaterSpec);
			glMaterialfv( GL_FRONT, GL_SHININESS, WaterShine);
			glNormal3f( 0.0f, 1.f, 0.0f);					
			glTexCoord2f(1.0f * 0.05f, 1.0f * 0.05f); 
			glVertex3f(-1000.0f, 0.0f, -1000.0f);
			glTexCoord2f(0.0f, 1.0f * 0.05f); 
			glVertex3f( 1000.0f, 0.0f, -1000.0f);
			glTexCoord2f(0.0f, 0.0f); 
			glVertex3f( 1000.0f, 0.0f, 1000.0f);
			glTexCoord2f(1.0f * 0.05f, 0.0f); 
			glVertex3f(-1000.0f, 0.0f, 1000.0f);
		}
		glEnd();

		//disable sphere mapping & alpha blending
		glDisable(GL_BLEND);
		glDisable(GL_TEXTURE_GEN_S);
		glDisable(GL_TEXTURE_GEN_T);
		glEnable(GL_CULL_FACE);
		glDepthMask(GL_TRUE);
	}
	glPopMatrix();
}

//Create terrain display list
void DEFAULT_TERRAIN::CreateTerrainDL()
{
	glNewList(TERRAIN_LIST, GL_COMPILE);
	{
		glPushMatrix();

		glBindTexture(GL_TEXTURE_2D, Textures[6]);
		
		//iterate thru the z axis of the plain, setting each vertex to the height 
		//of the corresponding greyscale color of the height map
		for(int z = 0; z <= m_Size; ++z)
		{
			glBegin(GL_TRIANGLES);
			{
				//iterate thru the x axis inorder to construct the triangle strips for the terrain mesh
				for(int x = 0; x <= m_Size; ++x)
				{
					//1
					int index = (z * m_Size) + x;
					int normalIndex = (z * m_Size) + x;
 
					glNormal3f(m_TerrainNormals[normalIndex].x, m_TerrainNormals[normalIndex].y, m_TerrainNormals[normalIndex].z);
					glTexCoord2f((float)x * 0.2f , (float)z * 0.2f);
					glVertex3f(x, m_heightData.m_Data[index] * m_heightScale, z);


					//2
					index = ((z+1) * m_Size) + x;
					normalIndex = ((z+1) * m_Size) + x;

					glNormal3f(m_TerrainNormals[normalIndex].x, m_TerrainNormals[normalIndex].y, m_TerrainNormals[normalIndex].z);
					glTexCoord2f((float)x * 0.2f, (float)(z + 1) * 0.2f);
					glVertex3f(x, m_heightData.m_Data[index] * m_heightScale, (z + 1));


					//3
					index = (z * m_Size + (x + 1));
					normalIndex = (z * m_Size + (x + 1));

					glNormal3f(m_TerrainNormals[normalIndex].x, m_TerrainNormals[normalIndex].y, m_TerrainNormals[normalIndex].z);
					glTexCoord2f((float)(x + 1) * 0.2f, (float)z * 0.2f);
					glVertex3f((x + 1), m_heightData.m_Data[index]* m_heightScale, z);

					//4
					index = (z * m_Size + (x + 1));
					normalIndex = (z * m_Size + (x + 1));

					glNormal3f(m_TerrainNormals[normalIndex].x, m_TerrainNormals[normalIndex].y, m_TerrainNormals[normalIndex].z);
					glTexCoord2f((float)(x + 1) * 0.2f, (float)z * 0.2f);
					glVertex3f((x + 1), m_heightData.m_Data[index]* m_heightScale, z);

					//5
					index = ((z+1) * m_Size) + x;
					normalIndex = ((z+1) * m_Size) + x;

					glNormal3f(m_TerrainNormals[normalIndex].x, m_TerrainNormals[normalIndex].y, m_TerrainNormals[normalIndex].z);
					glTexCoord2f((float)x * 0.2f, (float)(z + 1) * 0.2f);
					glVertex3f(x, m_heightData.m_Data[index]* m_heightScale, (z + 1));
	
					//6
					index = ((z+1) * m_Size) + (x + 1);
					normalIndex = ((z+1) * m_Size) + (x + 1);

					glNormal3f(m_TerrainNormals[normalIndex].x, m_TerrainNormals[normalIndex].y, m_TerrainNormals[normalIndex].z);
					glTexCoord2f((float)(x + 1) * 0.2f, (float)(z + 1) * 0.2f);
					glVertex3f((x + 1), m_heightData.m_Data[index]* m_heightScale, (z + 1));
					
					index++;
					normalIndex++;
				}
			}
			glEnd();
		}
		glPopMatrix();
	}
	glEndList();
}


//Initialize the terrain
void DEFAULT_TERRAIN::InitTerrain()
{
	//pre compute terrain normals here
	SetHeightScale(0.2f);
	ComputeNormals();
	ComputeVertexNormals();

	//create the terrain
	CreateTerrainDL();
}


//Default terrain rendering method implementation
void DEFAULT_TERRAIN::TerrainRender()
{
	//draw the skybox
	CreateSkyBox();

	//draw terrain
	glCallList(TERRAIN_LIST);

	//draw water
	CreateWater();
}